/*****************************************************************************
Copyright (C) 1993, 2001, 2010 Humboldt-Universitaet zu Berlin

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
/*** Beispiel 5: Transportsystem **************************/
#include <odemx/odemx.h>
using namespace odemx;
using namespace odemx::data;
using namespace odemx::base;
using namespace odemx::random;
using namespace odemx::synchronization;

#include	<iostream>
using namespace std;


Condq  *q;
Waitq  *ltq, *stq;
ContinuousDist  *nextl, *nexts;
class Sdigger *s;

class Truck : public Process {
public:
	double load;				// Fassungsvermoegen
	Truck(Simulation &sim, const Label &name) : Process(sim, name){}
};

class Digger : public Process {
public:
	Truck *t;				// zugeordneter Lastwagen
	double rate;				// Entladerate
	Digger(Simulation& sim, const Label &name) : Process(sim, name), t(0) {}
};

class Ltruck : public Truck {
public:
	Ltruck (Simulation& sim) : Truck(sim, "L"){}
	virtual int main();
};

class Struck : public Truck {
public:
	Struck(Simulation& sim) : Truck(sim, "S"){}
	virtual int main ();
};

class Ldigger : public Digger {
public:
	Ldigger(Simulation& sim) : Digger(sim, "Ldigger"){}
	virtual int main();
};

class Sdigger : public Digger {
public:
	SimTime started;				// Entladungsbeginn
	Sdigger(Simulation& sim) : Digger(sim, "Sdigger"){}
	virtual int main();
  bool cond () {
  // Aktivierungsbedingung f?r kleinen Bagger:
  // ein kleines Fahrzeug wartet und alle grossen Bagger sind
  // besetzt
    return
    ((stq->getWaitingSlaves ().size () > 0) || (ltq->getWaitingMasters ().size () == 0));
  }
};


class LtruckGenerator : public Process {
public:
	Ltruck *lt;
	LtruckGenerator(Simulation& sim) : Process(sim, "L-gen"){}
	virtual int main();
};

class StruckGenerator : public Process {
public:
	Struck *st;
	StruckGenerator(Simulation& sim) : Process(sim, "S-gen"){}
	virtual int main();
};

int Ltruck::main() {
	load = 20.0;
  while (load > 0.0) {
    q->signal(); // Aktivierung des kleinen Baggers
    ltq->wait(); // Warten auf beliebigen Bagger
  }
	return 0;
}

int LtruckGenerator::main() {
	for (;;) {
		lt = new Ltruck (getSimulation ());		// Generierung eines weiteren
		lt->activate () ;			// grossen Fahrzeugs mit
		holdFor (nextl->sample());		// Zwischenankunftszeit
	}
	return 0;
}

int Struck::main() {
	load = 5.0;
	if (s->getProcessState () == RUNNABLE && s->t) {
		if (dynamic_cast<Ltruck*>(s->t)) {
			// Unterbrechung des kleinen Baggers, der
			// gerade ein grosses Fahrzeug bedient
			s->interrupt ();
		}
		// Aktivierung des kleinen Baggers, falls er wartet
		else q->signal();
	}
	else q->signal();
	stq->wait();				// Warten auf kleinen Bagger
	return 0;
}

int StruckGenerator::main() {
	for (;;) {
		st = new Struck (getSimulation ());		// Generierung eines weiteren
		st->activate () ;			                // kleinen Fahrzeugs mit
		holdFor (nexts->sample());		        // Zwischenankunftszeit
	}
	return 0;
}

int Ldigger::main() {
	rate = 240.0;
	for (;;) {
	// Warten auf grosses Fahrzeug
		t = dynamic_cast<Truck*>(ltq->coopt());
	// Entladung
		holdFor  (t->load/rate);
		t->load = 0.0;
	// Freigabe des Fahrzeugs
		t->activate ();
		t = 0;
		if (s->t) {
			if (ltq->getWaitingSlaves ().size () == 0 && dynamic_cast<Ltruck*>(s->t)) {
			// Unterbrechung des kleinen Baggers, falls er
			// ein grosses Fahrzeug entlaedt
				s->interrupt ();
			}
		}
	}
	return 0;
}


int Sdigger::main() {
	rate = 60.0;
	for (;;) {
	// Warten auf Aktivierungsbedingung: Sdigger::cond()==TRUE
		q->wait((Condition) &Sdigger::cond);
		if (! stq->getWaitingSlaves ().empty ()) {
		// kleines Fahrzeug zur Entladung bereit
			t = dynamic_cast<Truck*>(stq->coopt());
			holdFor (t->load/rate);
		}
		else {
		// grosses Fahrzeug zur Entladung bereit
			started = getSimulation ().getTime ();
			t = dynamic_cast<Truck*>(ltq->coopt());
			holdFor  (t->load/rate);
			if (!isInterrupted ()) t->load = 0.0;
			else {
			// Unterbrechung der Entladung
			// beide Ursachen werden gleich behandelt
				t->load -= (getSimulation ().getTime () - started) * rate;
				t->setPriority(1);
				// resetInterrupt must be removed. is private
				//resetInterrupt ();
			}
		}
		// Freigabe des (teilweise) entladenen Fahrzeugs
		t->activate ();
		t = 0;
	}
	return 0;
}

int main () {
  Simulation& sim = getDefaultSimulation ();
	Ldigger           *ld;
	LtruckGenerator   *lt;
	StruckGenerator   *st;

  // Statistikpuffer
  using odemx::data::buffer::StatisticsBuffer;
  static std::tr1::shared_ptr < StatisticsBuffer > stats = 
    StatisticsBuffer::create (sim, "Statistik", false);

  sim.addConsumer (data::channel_id::statistics, stats);

  // Initialisierung
	nextl = new NegativeExponential(sim, "next large", 22.0);
	nexts = new NegativeExponential(sim, "next small", 10.0);
	q     = new Condq (sim, "sq");
	stq   = new Waitq (sim, "s_Truck_q");
	ltq   = new Waitq (sim, "l_Truck_q");
	s     = new Sdigger (sim);

	s->activate ();

	ld   = new Ldigger (sim); ld->activate ();
	ld   = new Ldigger (sim); ld->activate ();
	lt   = new LtruckGenerator (sim); lt->activate ();
	st   = new StruckGenerator (sim); st->activate ();

  sim.runUntil (10);

  // Report
  data::output::XmlReport xmlReport("digger.xml");
  xmlReport.addReportProducer (*stats);
  xmlReport.generateReport ();

  return 0;
}
